@@ -9,8 +9,13 @@ Features:
 * **Automatic offset correction** - Older VMAT tests had the ROIs offset, newer ones are centered. No worries, pylinac finds the ROIs automatically.
 * **Automatic open/DMLC identification** - Pass in both images--don't worry about naming. Pylinac will automatically identify the right images.
 """
+import dataclasses
+import enum
+import typing
+import webbrowser
+from dataclasses import dataclass
 from io import BytesIO
-from typing import Union, List, Tuple, Sequence, Optional
+from typing import Union, List, Tuple, Sequence, Optional, BinaryIO
 
 import argue
 import matplotlib.pyplot as plt
@@ -21,13 +26,43 @@ from .core.geometry import Point, Rectangle
 from .core.image import ImageLike
 from .core.io import get_url, TemporaryZipDirectory, retrieve_demo_file
 from .core.pdf import PylinacCanvas
-from .core.profile import SingleProfile
-from .core.utilities import open_path
+from .core.profile import SingleProfile, Interpolation, Edge
+from .core.utilities import ResultBase
 from .settings import get_dicom_cmap
 
-DMLC = 'dmlc'
-OPEN = 'open'
-PROFILE = 'profile'
+
+class ImageType(enum.Enum):
+    """The image type options"""
+
+    DMLC = "dmlc"  #:
+    OPEN = "open"  #:
+    PROFILE = "profile"  #:
+
+
+@dataclass
+class SegmentResult:
+    """An individual segment/ROI result"""
+
+    passed: bool  #:
+    x_position_mm: float  #:
+    r_corr: float  #:
+    r_dev: float  #:
+    center_x_y: float  #:
+
+
+@dataclass
+class VMATResult(ResultBase):
+    """This class should not be called directly. It is returned by the ``results_data()`` method.
+    It is a dataclass under the hood and thus comes with all the dunder magic.
+
+    Use the following attributes as normal class attributes."""
+
+    test_type: str  #:
+    tolerance_percent: float  #:
+    max_deviation_percent: float  #:
+    abs_mean_deviation: float  #:
+    passed: bool  #:
+    segment_data: typing.Iterable[SegmentResult]  #:
 
 
 class VMATBase:
@@ -40,7 +75,7 @@ class VMATBase:
     segments: List
     _tolerance: float
 
-    def __init__(self, image_paths: Sequence[str]):
+    def __init__(self, image_paths: Sequence[Union[str, BinaryIO]]):
         """
         Parameters
         ----------
@@ -83,11 +118,13 @@ class VMATBase:
     @classmethod
     def from_demo_images(cls):
         """Construct a VMAT instance using the demo images."""
-        demo_file = retrieve_demo_file(url=cls._url_suffix)
+        demo_file = retrieve_demo_file(name=cls._url_suffix)
         return cls.from_zip(demo_file)
 
     @argue.bounds(tolerance=(0, 8))
-    def analyze(self, tolerance: Union[float, int]=1.5):
+    def analyze(
+        self, tolerance: Union[float, int] = 1.5, segment_size_mm: Tuple = (5, 100)
+    ):
         """Analyze the open and DMLC field VMAT images, according to 1 of 2 possible tests.
 
         Parameters
@@ -95,15 +132,21 @@ class VMATBase:
         tolerance : float, int, optional
             The tolerance of the sample deviations in percent. Default is 1.5.
             Must be between 0 and 8.
+        segment_size_mm : tuple(int, int)
+            The (width, height) of the ROI segments in mm.
         """
-        self._tolerance = tolerance/100
+        self._tolerance = tolerance / 100
 
         """Analysis"""
         points = self._calculate_segment_centers()
+        Segment._nominal_width_mm = segment_size_mm[0]
+        Segment._nominal_height_mm = segment_size_mm[1]
         self._construct_segments(points)
 
     @staticmethod
-    def _load_images(image_paths: list) -> Tuple[ImageLike, ImageLike]:
+    def _load_images(
+        image_paths: Sequence[Union[str, BytesIO]]
+    ) -> Tuple[ImageLike, ImageLike]:
         image1 = image.load(image_paths[0])
         image2 = image.load(image_paths[1])
         image1.ground()
@@ -111,7 +154,9 @@ class VMATBase:
         return image1, image2
 
     @staticmethod
-    def _check_img_inversion(image1: ImageLike, image2: ImageLike) -> Tuple[ImageLike, ImageLike]:
+    def _check_img_inversion(
+        image1: ImageLike, image2: ImageLike
+    ) -> Tuple[ImageLike, ImageLike]:
         """Check that the images are correctly inverted."""
         for image in [image1, image2]:
             image.check_inversion()
@@ -120,8 +165,8 @@ class VMATBase:
     def _identify_images(self, image1: ImageLike, image2: ImageLike):
         """Identify which image is the DMLC and which is the open field."""
         profile1, profile2 = self._median_profiles((image1, image2))
-        field_profile1 = profile1.field_values()
-        field_profile2 = profile2.field_values()
+        field_profile1 = profile1.field_data()["field values"]
+        field_profile2 = profile2.field_data()["field values"]
         if np.std(field_profile1) > np.std(field_profile2):
             self.dmlc_image = image1
             self.open_image = image2
@@ -138,20 +183,47 @@ class VMATBase:
             The results string showing the overall result and deviation statistics by segment.
         """
         if self.passed:
-            passfail_str = 'PASS'
+            passfail_str = "PASS"
         else:
-            passfail_str = 'FAIL'
+            passfail_str = "FAIL"
 
-        string = f'{self._result_header}\nTest Results (Tol. +/-{self._tolerance*100:2.2}%): {passfail_str}\n'
+        string = f"{self._result_header}\nTest Results (Tol. +/-{self._tolerance*100:2.2}%): {passfail_str}\n"
 
-        string += f'Max Deviation: {self.max_r_deviation:2.3}%\nAbsolute Mean Deviation: {self.avg_abs_r_deviation:2.3}%'
+        string += f"Max Deviation: {self.max_r_deviation:2.3}%\nAbsolute Mean Deviation: {self.avg_abs_r_deviation:2.3}%"
         return string
 
+    def results_data(self, as_dict=False) -> Union[VMATResult, dict]:
+        """Present the results data and metadata as a dataclass or dict.
+        The default return type is a dataclass."""
+        segment_data = []
+        for idx, segment in enumerate(self.segments):
+            segment_data.append(
+                SegmentResult(
+                    passed=segment.passed,
+                    r_corr=segment.r_corr,
+                    r_dev=segment.r_dev,
+                    center_x_y=segment.center.as_array(),
+                    x_position_mm=self.SEGMENT_X_POSITIONS_MM[idx],
+                )
+            )
+        data = VMATResult(
+            test_type=self._result_header,
+            tolerance_percent=self._tolerance * 100,
+            max_deviation_percent=self.max_r_deviation,
+            abs_mean_deviation=self.avg_abs_r_deviation,
+            passed=self.passed,
+            segment_data=segment_data,
+        )
+
+        if as_dict:
+            return dataclasses.asdict(data)
+        return data
+
     def _calculate_segment_centers(self) -> List[Point]:
         """Construct the center points of the segments based on the field center and known x-offsets."""
         points = []
         dmlc_prof, _ = self._median_profiles((self.dmlc_image, self.open_image))
-        x_field_center, _ = dmlc_prof.fwxm_center()
+        x_field_center = dmlc_prof.beam_center()["index (rounded)"]
         for x_offset_mm in self.SEGMENT_X_POSITIONS_MM:
             y = self.open_image.center.y
             x_offset_pixels = x_offset_mm * self.open_image.dpmm
@@ -196,7 +268,7 @@ class VMATBase:
         """Return the value of the maximum R_deviation segment."""
         return np.max(np.abs(self.r_devs))
 
-    def plot_analyzed_image(self, show: bool=True):
+    def plot_analyzed_image(self, show: bool = True, **plt_kwargs: dict):
         """Plot the analyzed images. Shows the open and dmlc images with the segments drawn; also plots the median
         profiles of the two images for visual comparison.
 
@@ -204,22 +276,25 @@ class VMATBase:
         ----------
         show : bool
             Whether to actually show the image.
+        plt_kwargs : dict
+            Keyword args passed to the plt.subplots() method. Allows one to set things like figure size.
         """
-        fig, axes = plt.subplots(ncols=3, sharex=True)
-        subimages = (OPEN, DMLC, PROFILE)
-        titles = ('Open', 'DMLC', 'Median Profiles')
+        fig, axes = plt.subplots(ncols=3, sharex=True, **plt_kwargs)
+        subimages = (ImageType.OPEN, ImageType.DMLC, ImageType.PROFILE)
+        titles = ("Open", "DMLC", "Median Profiles")
         for subimage, axis, title in zip(subimages, axes, titles):
             self._plot_analyzed_subimage(subimage=subimage, ax=axis, show=False)
             axis.set_title(title)
-        axis.set_ylabel('Normalized Response')
-        axis.legend(loc='lower center')
+        axis.set_ylabel("Normalized Response")
+        axis.legend(loc="lower center")
 
         if show:
             plt.tight_layout(h_pad=1.5)
             plt.show()
 
-    @argue.options(subimage=(DMLC, OPEN, PROFILE))
-    def _save_analyzed_subimage(self, filename: str, subimage: str, **kwargs):
+    def _save_analyzed_subimage(
+        self, filename: Union[str, BytesIO], subimage: ImageType, **kwargs
+    ):
         """Save the analyzed images as a png file.
 
         Parameters
@@ -232,8 +307,9 @@ class VMATBase:
         self._plot_analyzed_subimage(subimage=subimage, show=False)
         plt.savefig(filename, **kwargs)
 
-    @argue.options(subimage=(DMLC, OPEN, PROFILE))
-    def _plot_analyzed_subimage(self, subimage: str, show: bool=True, ax: Optional[plt.Axes]=None):
+    def _plot_analyzed_subimage(
+        self, subimage: ImageType, show: bool = True, ax: Optional[plt.Axes] = None
+    ):
         """Plot an individual piece of the VMAT analysis.
 
         Parameters
@@ -250,24 +326,26 @@ class VMATBase:
             fig, ax = plt.subplots()
 
         # plot DMLC or OPEN image
-        if subimage in (DMLC, OPEN):
-            if subimage == DMLC:
+        if subimage in (ImageType.DMLC, ImageType.OPEN):
+            if subimage == ImageType.DMLC:
                 img = self.dmlc_image
-            elif subimage == OPEN:
+            elif subimage == ImageType.OPEN:
                 img = self.open_image
             ax.imshow(img, cmap=get_dicom_cmap())
             self._draw_segments(ax)
             plt.sca(ax)
-            plt.axis('off')
+            plt.axis("off")
             plt.tight_layout()
 
         # plot profile
-        elif subimage == PROFILE:
-            dmlc_prof, open_prof = self._median_profiles((self.dmlc_image, self.open_image))
-            ax.plot(dmlc_prof.values, label='DMLC')
-            ax.plot(open_prof.values, label='Open')
-            ax.autoscale(axis='x', tight=True)
-            ax.legend(loc=8, fontsize='large')
+        elif subimage == ImageType.PROFILE:
+            dmlc_prof, open_prof = self._median_profiles(
+                (self.dmlc_image, self.open_image)
+            )
+            ax.plot(dmlc_prof.values, label="DMLC")
+            ax.plot(open_prof.values, label="Open")
+            ax.autoscale(axis="x", tight=True)
+            ax.legend(loc=8, fontsize="large")
             ax.grid()
 
         if show:
@@ -288,9 +366,17 @@ class VMATBase:
     @staticmethod
     def _median_profiles(images) -> Tuple[SingleProfile, SingleProfile]:
         """Return two median profiles from the open and dmlc image. For visual comparison."""
-        profile1 = SingleProfile(np.mean(images[0], axis=0))
+        profile1 = SingleProfile(
+            np.mean(images[0], axis=0),
+            interpolation=Interpolation.NONE,
+            edge_detection_method=Edge.INFLECTION_DERIVATIVE,
+        )
         profile1.stretch()
-        profile2 = SingleProfile(np.mean(images[1], axis=0))
+        profile2 = SingleProfile(
+            np.mean(images[1], axis=0),
+            interpolation=Interpolation.NONE,
+            edge_detection_method=Edge.INFLECTION_DERIVATIVE,
+        )
         profile2.stretch()
 
         # normalize the profiles to approximately the same value
@@ -301,7 +387,13 @@ class VMATBase:
 
         return profile1, profile2
 
-    def publish_pdf(self, filename: str, notes: str=None, open_file: bool=False, metadata: Optional[dict]=None):
+    def publish_pdf(
+        self,
+        filename: str,
+        notes: str = None,
+        open_file: bool = False,
+        metadata: Optional[dict] = None,
+    ):
         """Publish (print) a PDF containing the analysis, images, and quantitative results.
 
         Parameters
@@ -321,23 +413,33 @@ class VMATBase:
             Unit: TrueBeam
             --------------
         """
-        canvas = PylinacCanvas(filename=filename, page_title=f"{self._result_short_header} VMAT Analysis", metadata=metadata)
-        for y, x, width, img in zip((9, 9, -2), (1, 11, 3), (9, 9, 14), (OPEN, DMLC, PROFILE)):
+        canvas = PylinacCanvas(
+            filename=filename,
+            page_title=f"{self._result_short_header} VMAT Analysis",
+            metadata=metadata,
+        )
+        for y, x, width, img in zip(
+            (9, 9, -2),
+            (1, 11, 3),
+            (9, 9, 14),
+            (ImageType.OPEN, ImageType.DMLC, ImageType.PROFILE),
+        ):
             data = BytesIO()
             self._save_analyzed_subimage(data, subimage=img)
             canvas.add_image(data, location=(x, y), dimensions=(width, 18))
             # canvas.add_text(text=f"{img} Image", location=(x + 2, y + 10), font_size=18)
-        canvas.add_text(text='Open Image', location=(4, 22), font_size=18)
-        canvas.add_text(text=f'{self.open_image.base_path}', location=(4, 21.5))
-        canvas.add_text(text='DMLC Image', location=(14, 22), font_size=18)
-        canvas.add_text(text=f'{self.dmlc_image.base_path}', location=(14, 21.5))
-        canvas.add_text(text='Median profiles', location=(8, 12), font_size=18)
-        text = [f'{self._result_header} VMAT results:',
-                f'Source-to-Image Distance (mm): {self.open_image.sid:2.0f}',
-                f'Tolerance (%): {self._tolerance*100:2.1f}',
-                f'Absolute mean deviation (%): {self.avg_abs_r_deviation:2.2f}',
-                f'Maximum deviation (%): {self.max_r_deviation:2.2f}',
-                ]
+        canvas.add_text(text="Open Image", location=(4, 22), font_size=18)
+        canvas.add_text(text=f"{self.open_image.base_path}", location=(4, 21.5))
+        canvas.add_text(text="DMLC Image", location=(14, 22), font_size=18)
+        canvas.add_text(text=f"{self.dmlc_image.base_path}", location=(14, 21.5))
+        canvas.add_text(text="Median profiles", location=(8, 12), font_size=18)
+        text = [
+            f"{self._result_header} VMAT results:",
+            f"Source-to-Image Distance (mm): {self.open_image.sid:2.0f}",
+            f"Tolerance (%): {self._tolerance*100:2.1f}",
+            f"Absolute mean deviation (%): {self.avg_abs_r_deviation:2.2f}",
+            f"Maximum deviation (%): {self.max_r_deviation:2.2f}",
+        ]
         canvas.add_text(text=text, location=(10, 25.5))
         if notes is not None:
             canvas.add_text(text="Notes:", location=(1, 5.5), font_size=14)
@@ -346,14 +448,15 @@ class VMATBase:
         canvas.finish()
 
         if open_file:
-            open_path(filename)
+            webbrowser.open(filename)
 
 
 class DRGS(VMATBase):
     """Class representing a Dose-Rate, Gantry-speed VMAT test. Will accept, analyze, and return the results."""
-    _url_suffix = 'drgs.zip'
-    _result_header = 'Dose Rate & Gantry Speed'
-    _result_short_header = 'DR/GS'
+
+    _url_suffix = "drgs.zip"
+    _result_header = "Dose Rate & Gantry Speed"
+    _result_short_header = "DR/GS"
     SEGMENT_X_POSITIONS_MM = (-60, -40, -20, 0, 20, 40, 60)
 
     @staticmethod
@@ -367,9 +470,10 @@ class DRGS(VMATBase):
 
 class DRMLC(VMATBase):
     """Class representing a Dose-Rate, MLC speed VMAT test. Will accept, analyze, and return the results."""
-    _url_suffix = 'drmlc.zip'
-    _result_header = 'Dose Rate & MLC Speed'
-    _result_short_header = 'DR/MLCS'
+
+    _url_suffix = "drmlc.zip"
+    _result_header = "Dose Rate & MLC Speed"
+    _result_short_header = "DR/MLCS"
     SEGMENT_X_POSITIONS_MM = (-45, -15, 15, 45)
 
     @staticmethod
@@ -396,12 +500,18 @@ class Segment(Rectangle):
     passed : boolean
         Specifies where the segment reading deviation was under tolerance.
     """
-    # width of the segment (i.e. parallel to MLC motion) in pixels under reference conditions
-    _nominal_width_mm: int = 5
-    _nominal_height_mm: int = 100
 
-    def __init__(self, center_point: Point, open_image: image.DicomImage, dmlc_image: image.DicomImage,
-                 tolerance: Union[float, int]):
+    # width of the segment (i.e. parallel to MLC motion) in pixels under reference conditions
+    _nominal_width_mm: int
+    _nominal_height_mm: int
+
+    def __init__(
+        self,
+        center_point: Point,
+        open_image: image.DicomImage,
+        dmlc_image: image.DicomImage,
+        tolerance: Union[float, int],
+    ):
         self.r_dev: float = 0.0  # is assigned after all segments constructed
         self._tolerance = tolerance
         self._open_image = open_image
@@ -413,10 +523,14 @@ class Segment(Rectangle):
     @property
     def r_corr(self) -> float:
         """Return the ratio of the mean pixel values of DMLC/OPEN images."""
-        dmlc_value = self._dmlc_image.array[self.bl_corner.y:self.bl_corner.y + self.height,
-                     self.bl_corner.x: self.bl_corner.x + self.width].mean()
-        open_value = self._open_image.array[self.bl_corner.y:self.bl_corner.y + self.height,
-                     self.bl_corner.x: self.bl_corner.x + self.width].mean()
+        dmlc_value = self._dmlc_image.array[
+            self.bl_corner.y : self.bl_corner.y + self.height,
+            self.bl_corner.x : self.bl_corner.x + self.width,
+        ].mean()
+        open_value = self._open_image.array[
+            self.bl_corner.y : self.bl_corner.y + self.height,
+            self.bl_corner.x : self.bl_corner.x + self.width,
+        ].mean()
         ratio = (dmlc_value / open_value) * 100
         return ratio
 
@@ -427,4 +541,4 @@ class Segment(Rectangle):
 
     def get_bg_color(self) -> str:
         """Get the background color of the segment when plotted, based on the pass/fail status."""
-        return 'blue' if self.passed else 'red'
+        return "blue" if self.passed else "red"
